"""
A set of methods for reading data files.

Copyright 2015, University of Freiburg.

Elmar Haussmann <haussmann@cs.uni-freiburg.de>
"""
__author__ = 'haussmae'
import globals
import re
import logging


logger = logging.getLogger(__name__)


def read_relation_expected_types(target_types_file):
    relation_target_types = {}
    with open(target_types_file, 'r') as f:
        for line in f:
            cols = line.strip().split('\t')
            relation_name = cols[0]
            type_name = cols[1]
            relation_target_types[relation_name] = type_name
    return relation_target_types


def read_relation_lemmas(lemma_file):
    """Read the lemmatized version of each relation.

    :param lemma_file:
    :return:
    """
    logger.info("Reading relation lemmas from %s." % lemma_file)
    relation_lemma = dict()
    with open(lemma_file, 'r') as f:
        for line in f:
            cols = line.decode('utf-8').strip().split('\t')
            name = cols[0]
            lemma_name = cols[1]
            relation_lemma[name] = lemma_name
    logger.info("Read %s relation lemmas." % len(relation_lemma))
    return relation_lemma


def read_word_type_distributions(word_type_counts_file,
                                 min_count=200,
                                 min_type_count=10):
    """Read the word entity type counts.

    The filtering parameters are really only for keeping the data footprint
    small and (should/are meant to) have no other effect.
    :param word_type_counts_file:
    :param min_count: The minimum number of times we must have seen
    the word to consider it.
    :return:
    """
    word_type_counts = dict()
    logger.info("Reading word type counts from %s." % word_type_counts_file)
    with open(word_type_counts_file, 'r') as f:
        for line in f:
            cols = line.decode('utf-8').strip().split('\t')
            word = cols[0]
            word_total_count = int(cols[1])
            if word_total_count < min_count:
                continue
            types = cols[2].split(' ')
            word_type_counts[word] = dict()
            for t in types:
                ind = t.rindex(':')
                type_name = t[:ind]
                count = int(t[ind + 1:])
                if count < min_type_count:
                    continue
                word_type_counts[word][type_name] = count
    logger.info("Read type counts for %s words." % len(word_type_counts))
    return word_type_counts


def read_word_derivations(derivations_file):
    """Read the word derivations from a file.

    Returns a dictionary from word to a set of words which are its
    derivations.
    :param derivations_file:
    :return:
    """
    word_derivations = dict()
    logger.info("Reading word derivations from %s." % derivations_file)
    with open(derivations_file, "r") as f:
        for line in f:
            cols = line.strip().decode('utf-8').split('\t')
            word = cols[0]
            derivations = set(cols[1].split(' '))
            word_derivations[word] = derivations
    logger.info("Read derivations for %s words." % len(word_derivations))
    return word_derivations


def read_relation_target_type_distributions(target_types_file):
    relation_target_types = {}
    ignore_types = {'common.topic'}
    with open(target_types_file, 'r') as f:
        for line in f:
            cols = line.strip().split('\t')
            relation_name = cols[0]
            count = int(cols[1])
            types = cols[2].split(' ')
            relation_target_types[relation_name] = dict()
            for t in types:
                if t in ignore_types:
                    continue
                type_name = t[:t.rindex(':')]
                if type_name in ignore_types:
                    continue
                type_count = float(t[t.rindex(':') + 1:])
                relation_target_types[relation_name][type_name] = type_count
            # Avoid empty dicts if all types were ignored.
            if not relation_target_types[relation_name]:
                del relation_target_types[relation_name]
    return relation_target_types


def read_relation_words(relation_words_file, n_top_words=30):
    """
    Read the top relation indicator words from file.
    Return a map from relation -> {word -> score}
    :param relation_words_file:
    :return:
    """
    word_score_re = re.compile(r'([^(]+)\(([^\)]+)\)')
    relation_words = {}
    logging.info("Reading relation words from %s ." % relation_words_file)
    with open(relation_words_file, 'r') as f:
        for line in f:
            cols = line.strip().decode('utf-8').split('\t')
            relation_name = cols[0]
            top_words = []
            words = cols[2].split(' ')
            for c in words[:n_top_words]:
                m = re.match(word_score_re, c)
                if m:
                    word = m.group(1)
                    score = float(m.group(2))
                    top_words.append((word, score))
                else:
                    logger.warn("Unknown word format in "
                                "relation words list: %s." % line)
            # Normalize the scores.
            top_words_dict = normalize_word_scores(top_words)
            relation_words[relation_name] = top_words_dict
    logging.info("Read words for %s relations." % len(relation_words))
    return relation_words


def normalize_word_scores(word_scores):
    """
    Returns a dict from word -> normalized score
    :param word_scores: A list of tuples (word, score)
    :return:
    """
    total_score = float(sum([s for w, s in word_scores]))
    return {w: s / total_score for w, s in word_scores}


def read_mediated_relation_words(relation_words_file, n_top_words=10):
    """
    Read the top relation indicator words from file.
    Return a map from relation -> {word -> score}
    :param relation_words_file:
    :return:
    """
    word_score_re = re.compile(r'([^(]+)\(([^\)]+)\)')
    relation_words = {}
    logging.info("Reading mediated relation words from %s ." %
                 relation_words_file)
    with open(relation_words_file, 'r') as f:
        for line in f:
            cols = line.strip().decode('utf-8').split('\t')
            relation_names = tuple(cols[0].split(' '))
            top_words = []
            words = cols[2].split(' ')
            for c in words[:n_top_words]:
                m = re.match(word_score_re, c)
                if m:
                    word = m.group(1)
                    score = float(m.group(2))
                    top_words.append((word, score))
            top_words_dict = normalize_word_scores(top_words)
            relation_words[relation_names] = top_words_dict
    logging.info("Read words for %s mediated relations." % len(relation_words))
    return relation_words


def read_reverse_relations(reverse_relations_file):
    """
    Read the reverse relations file and return a dict.
    :param reverse_relations_file:
    :return:
    """
    prefix = "<http://rdf.freebase.com/ns"
    reverse_relations = {}
    with open(reverse_relations_file, 'r') as f:
        for line in f:
            cols = line.strip().decode('utf-8').split('\t')
            name = cols[0]
            reverse = cols[1]
            # name = globals.get_mid_from_qualified_string(cols[0])
            # reverse = globals.get_mid_from_qualified_string(cols[1])
            # name = name.replace('/', '.')
            # reverse = reverse.replace('/', '.')
            reverse_relations[name] = reverse
            reverse_relations[reverse] = name
    return reverse_relations


def read_relation_counts(name_mapping_file):
    """
    Read the relation counts file and return a dict from
    relation -> (total, unique_subjects, unique_objects)
    :param name_mapping_file:
    :return:
    """
    relation_counts = {}
    logger.info("Reading relation counts from %s." % name_mapping_file)
    with open(name_mapping_file, 'r') as f:
        for line in f:
            cols = line.strip().decode('utf-8').split('\t')
            name = cols[0]
            # Some weird 1-character relations...
            if len(name) > 1:
                # Ignore counts of key relations
                if name.startswith(globals.FREEBASE_KEY_PREFIX):
                    continue
                # name = name[1:].replace('/', '.')
                counts = (int(cols[1]), int(cols[2]), int(cols[3]))
                relation_counts[name] = counts
    logger.info("Read relation counts for %s relations." %
                len(relation_counts))
    return relation_counts


def read_mediator_names(mediator_names_file):
    """
    Read the mediator names file and return a map from
    mid to name.
    :param name_mapping_file:
    :return:
    """
    mediator_names = {}
    with open(mediator_names_file, 'r') as f:
        for line in f:
            line = line.decode('utf-8')
            mid, name = line.strip().split('\t')
            mediator_names[name] = mid
    return mediator_names


def read_mediator_relations(mediator_relations_file):
    """
    Read the mediator names file and return a set mediator
    relations.
    :param name_mapping_file:
    :return:
    """
    mediator_relations = set()
    with open(mediator_relations_file, 'r') as f:
        for line in f:
            relation = line.decode('utf-8').strip()
            mediator_relations.add(relation)
    return mediator_relations


def read_temporal_predicate_from(temporal_predicate_pairs_file):
    """Read the temporal predicate pairs and return a dictionary .

    :param temporal_predicate_pairs_file:
    :return:
    """
    logger.info("Reading temporal predicate pairs from %s." % temporal_predicate_pairs_file)
    temporal_predicate_from = {}
    with open(temporal_predicate_pairs_file, 'r') as f:
        for line in f:
            cols = line.decode('utf-8').strip().split('\t')
            date_from = cols[0]
            date_to = cols[1]
            temporal_predicate_from[date_from] = date_to
        #    temporal_predicate_from[date_to] = date_to
    logger.info("Read %s temporal predicate pairs." % len(temporal_predicate_from))
    return temporal_predicate_from

def read_temporal_predicate_to(temporal_predicate_pairs_file):
    """Read the temporal predicate pairs and return a dictionary . 

    :param temporal_predicate_pairs_file:
    :return:
    """
    logger.info("Reading temporal predicate pairs from %s." % temporal_predicate_pairs_file)
    temporal_predicate_to = {}
    with open(temporal_predicate_pairs_file, 'r') as f:
        for line in f:
            cols = line.decode('utf-8').strip().split('\t')
            date_from = cols[0]
            date_to = cols[1]
            temporal_predicate_to[date_to] = date_from
       #     temporal_predicate_to[date_from] = date_from
    logger.info("Read %s temporal predicate pairs." % len(temporal_predicate_to))
    return temporal_predicate_to


def read_relation_expected_types(relation_expected_types_file):
    """Read the temporal predicate pairs and return a dictionary .

    :param temporal_predicate_pairs_file:
    :return:
    """
    logger.info("Reading relation expected types from %s." % relation_expected_types_file)
    relation_expected_types = {}
    with open(relation_expected_types_file, 'r') as f:
        for line in f:
            cols = line.decode('utf-8').strip().split('\t')
            relation = cols[0]
            expected_types = cols[1]
            relation_expected_types[relation] = expected_types
       #     temporal_predicate_to[date_from] = date_from
    logger.info("Read %s temporal predicate pairs." % len(relation_expected_types))
    return relation_expected_types

def read_stop_words(stop_words_file):
    """Read the stop word file and return a stop word list 

    :param stop_words_file:
    :return:
    """
    logger.info("Reading stop word file from %s." % stop_words_file)
    stop_words = []
    with open(stop_words_file) as f:
    
        for line in f:
            word = line.rstrip()
            if word not in stop_words:
                stop_words.append(word)
                print word
    
    logger.info("Read %s stop words." % len(stop_words))
    return stop_words
   
